default_hparams_dict = {
	

    "seed":11,
    
    ##################################################################
    #  Methods 
    ##################################################################
    # 
    # method (string):
    #       One of ['advi, 'gaussian, 'flows', 'meanfield', 'custom']
    #       choosing a method gives your some default hyper-parameter settings
    #       choosing a method will change the values for other relevant 
    #       hyper-parameters; however, choosing 'custom', will leave them as is.
    "method":"meanfield",

    ##################################################################
    #  ADVI hparams
    ##################################################################
    #
    # advi_use (bool):
    #       If true, alerts the internal functions that you are running 
    #       ADVI as reported in paper https://www.jmlr.org/papers/volume18/16-107/16-107.pdf.
    #       You can custom use different functionalities of ADVI implementation; however, you
    #       would want to switch 'advi_use' to False in that case. 
    # advi_convergence_threshold (float):
    #       The threshold parameters is used to diagnose convergence. If the 
    #       mean or median of the relative tolerance window is below the 
    #       threshold parameter, then the optimization is stopped.  
    # advi_step_size (float):
    #       parameter eta as defined in the original paper.
    # advi_adapt_step_size (bool):
    #       Default is True. If True, then it uses the heuristic step-size scheme in the 
    #       PyStan's ADVI implementation. 
    # advi_adapt_step_size_range (iterable):t
    #       The range used in the original paper was [100,10,1,0.1,0.01] 
    # advi_adapt_step_size_num_iters (int):
    #       The number of iterations to use for each step-size candidate, ADVI greedily chooses
    #       the maximum step-size based on the final ELBO value after 'advi_adapt_step_size_num_iters'
    #       iterations.
    # advi_callback_iterations (int):
    #       # of iterations after which ELBO is evaluated. This evaluation is then used to monitor 
    #       convergence by monitoring the mean and the median of relative tolerance within a 
    #       window of values. The length of this window (circular buffer) is heuristically determined. 

    "advi_use":False,
    'advi_convergence_threshold':0.001, 
    "advi_step_size":0.01,
    "advi_adapt_step_size":True,
    "advi_adapt_step_size_range":[0.1, 0.01],
    'advi_adapt_step_size_verbose':False, 
    'advi_adapt_step_size_num_iters':20, 
    'advi_callback_iteration':2,

    ##################################################################
    #  Optimization
    ##################################################################
    # step_size_base (float):
    #       Default value is 0.1. 
    #
    # step_size_scale (float):
    #       Default value is 4.0 
    #
    # step_size_exp (int):
    #       Default value is 0 
    #
    # step_size (float):
    #       Default value is 0.01 
    #       If comprehensive_step_search_scaling is on, than this overriden by
    #       step_size = (step_size_base/(step_size_scale**step_size_exp))/# of z_dims
    #
    # step_size_exp_range (iterable):
    #       Not supported right now.
    #       Default value is [0,1,2,3]
    #       For comprehensive step-search, we optimize for different step_sizes,
    #       where the different step_sizes are generated using step_size_exp_range.
    #       Comprehensive step-search is not supported right now.
    #
    # max_iters (int):
    #       # of optimization iterations. 
    #       In case of ADVI, if we choose to adapt step-size, then this is only used
    #       for final adapted step_size. 
    #
    # optimizer (string):
    #       One of ['adam', 'sgd', 'rmsprop', 'advi']
    #       'advi' is the custom step-size sequence proposed in the ADVI paper
    #       (see https://www.jmlr.org/papers/volume18/16-107/16-107.pdf page 12.)
    #       Currently, only 'advi' and 'adam' support early exit based on the callback function
    #
    # M_iw_train (int):
    #       # of importance weights during training to optimize the IW-ELBO.
    #       Setting M_iw_train=1 reverts to naive VI (optimizer ELBO).
    #
    # grad_estimator (string):
    #       One of ['DReG', 'STL', 'Total-gradient', 'closed-form-entropy']
    #       Total gradient refers to the regular IW-ELBO gradient. With M_iw_train = 1
    #       this would be same as regular VI. 'closed-form-entropy' works when
    #       families support closed-form-entropy calculation ("gaussian" or "diagonal".)
    #
    # per_iter_sample_budget (int):
    #       Fixes the # of samples used at each iteration to calculate the gradient and the 
    #       evaluation_fn. If per_iter_sample_budget = 100 and M_iw_train = 10, then we
    #       use 10 copies of IW-ELBO gradient to optimize at each iteration.
    #
    # comprehensive_step_search_scaling (bool):
    #       Default is False. If True, will scale the step-size using the heuristic used 
    #       in the paper https://arxiv.org/pdf/2006.10343.pdf. If False, it will use the 'step_size'
    #       as is. 

    "step_size":0.01,
    "comprehensive_step_search_scaling":False,
    "step_size_base":0.01,
    "step_size_scale":4.0,
    "step_size_exp":0, 
    "step_size_exp_range":[0,1,2,3], 
    "max_iters":100,
    "optimizer":"adam" ,
    "M_iw_train":1, 
    "grad_estimator":"DReG", 
    "per_iter_sample_budget":100, 

    ##################################################################
    #  Evaluation Function
    ##################################################################
    #
    # evaluation_fn (string):
    #       One of ['IWELBO', 'ELBO-cfe']
    #       When using different gradient estimators like DREG and STL, the objective 
    #       function is a surrogate function that has the same gradient as IW-ELBO
    #       but is different in value. evaluation_fn helps you monitor the correct 
    #       objective. 
    #       If set to IW-ELBO, it uses the same M as M_iw_train
    #       If set to ELBO-cfe, it monitors the ELBO using the closed-form entropy function

    "evaluation_fn":"IWELBO",

    ##################################################################
    #  VI Families
    ##################################################################
    #
    #  "vi_family" (string):
    #       One of ["rnvp", "gaussian", "diagonal"]

    "vi_family":"gaussian",

    ##################################################################
    #  RealNVP
    ##################################################################
    #
    # rnvp_num_transformations (int):
    #       # of coupling transforms; each transforms includes two transitions
    #
    # rnvp_num_hidden_layers (int):
    #       # of hidden layers within s-t block
    #
    # rnvp_num_hidden_units (int):
    #       # of units in each hidden layer
    #
    # params_init_scale (int):
    #       scaling to help initializes to standard normal approximately

    "rnvp_num_transformations":10,
    "rnvp_num_hidden_units":16,
    "rnvp_num_hidden_layers":2,
    "rnvp_params_init_scale":0.01,


    ##################################################################
    #  LI:Laplace's Initialization
    ##################################################################
    #
    # LI (bool):
    #       Default is False. If True, initializes the Gaussian parameters to 
    #       Laplace's approximation. Do not use LI with real-NVP
    #
    # LI_max_iters (int):
    #       Default is 2000. # of iterations for MAP estimate used by scipy.optimize.minimize
    #
    # LI_epsilon (float):
    #       Default is 1e-6. A small positive used to calculate Hessian at MAP
    #       estimate using finite differences.

    "LI":False,
    "LI_max_iters":2000,
    "LI_epsilon":1e-6,

    ##################################################################
    #  Other flexibilities
    ##################################################################
    #
    # fix_sample_budget (bool):
    #       Default is False. If False, will increase the per_iter_sample budget
    #       by M_iw_train times as this sets the 'num_copies_training' parameter
    #       to be same as 'per_iter_sample_budget'.  The final number of samples at 
    #       each iteration is (num_copies_training, M_iw_train). If True, this will 
    #       set num_copies_train = per_iter_sample_budget//M_iw_train
    #
    "fix_sample_budget":False,
}


import pystan
import warnings
import pickle
import os
import collections
import scipy
import warnings
import traceback
import functools
import autograd
import autograd.numpy as np
import vistan.utilities as utils
import logging
# get pystan to stop warning
logging.getLogger("pystan").propagate = False
###########################################################################
#  stuff related to loading stan models and data
###########################################################################


def is_good_model(code, data, model_name="test_model", verbose=True):
    """ A simple function to test if PyStan compiles, samples, optimizes
        the model correctly or not.

    Args:
        code (string): Stan code
        data (dict): Observations in the form of a dictionary
        model_name (str, optional): Defaults to "test_model".
        verbose (bool, optional): Defaults to True.

    Raises:
        ValueError: NUTS and ADVI unconstrained results do not match in
                    # of dimensions.
        ValueError: NUTS and Optimization results do not match in # of
                    dimensions.

    Returns:
        bool: Test result
    """

    if model_name is not None:
        model_name = model_name.replace("-", "_")
    try:

        model = Model(code, data, model_name, verbose=verbose)
        Z_nuts = model.sampling(iter=20, verbose=verbose)
        Z_advi = model.advi(iter=20, verbose=verbose)
        for Z in [Z_advi]:
            if Z.shape[1] != Z_nuts.shape[1]:
                raise ValueError("NUTS and VI not same dimensions")

        # also check that optimization works
        z, rez = model.argmax(
                            True, method='BFGS',
                            gtol=1e-3, maxiter=20)

        if len(z) != Z_nuts.shape[1]:
            raise ValueError("NUTS and argmax not same dimensions")

        return True
    except Exception as e:
        print(f"Error during model check...")
        raise e

###########################################################################
#  Compiling and Caching Stan models
###########################################################################


def get_compiled_model(model_code, model_name=None, verbose=False, **kwargs):
    """ A function to get PyStan compiled model.

    Args:
        model_code (string):
            Stan code
        model_name (string, optional):
            Defaults to None.
        verbose (bool, optional):
            Helps print some PyStan compilation warnings.
            Defaults to False.

    Returns:
        StanModel: Compiled StanModel for the model specified by model_code
    """
    if not os.path.exists('data/cached-models'):
        print('creating cached-models/ to save compiled stan models')
        os.makedirs('data/cached-models')

    cache_fn = f'data/cached-models/\
                        {utils.get_cache_fname(model_name, model_code)}.pkl'

    if os.path.isfile(cache_fn):
        try:
            with open(cache_fn, 'rb') as f:
                sm = pickle.load(f)
            print("Compiled model found...")
        except Exception as e:
            print("Error during re-loading the complied model.")
            print(f"Try recompiling the model. Changed the\
                    name of the model or delete \
                    the saved pickled file at {cache_fn}.")
            raise e
    else:
        try:
            print("Cached model not found. Compiling...")
            with utils.suppress_stdout_stderr(verbose):
                sm = pystan.StanModel(
                                    model_code=model_code,
                                    model_name=model_name, **kwargs)
        except Exception as e:
            print("Error during compilation...")
            print('Could not compile code for ' + model_code)
            raise e

        print("Caching model...")
        with open(cache_fn, 'wb') as f:
            pickle.dump(sm, f)
    return sm

###########################################################################
#  Class to provide the log_prob function that hooks into autograd
###########################################################################


class Model:
    def __init__(self, model_code, data, model_name=None, verbose=False):
        """
            A class to interface with the autograd.

            Args
            ----------
                model_code (string):
                    Stan code for the model
                data (dict):
                    Data in the dictionary format
                model_name (string):
                    Name of the model for easier identification.
                    This along with code is used to cache compiled models.

                verbose (bool):
                    If True, it will print additional details involving
                    Stan compilation.

            Attributes
            ----------
                data(dict):
                     Model data
                model_name(string):
                     If None, only model_code is used to cache.
                model_code(string):
                     Stan code. Also, used to cache.
                sm(StanModel):
                      Complied StanModel instance.
                fit(StandModelFit4):
                    A StanModelFit4 instance obtained using self.sm.sampling
                keys(list):
                     Names of the unconstrained parameters.
                zlen(list):
                     Number of latent dimensions in the model.

        """
        # THIS IS SLOW! TURN OPTIMIZATION ON SOMEDAY!
        extra_compile_args = ['-O1', '-w', '-Wno-deprecated']

        self.data = data
        self.model_name = model_name
        self.model_code = model_code

        self.sm = get_compiled_model(
                        model_code=self.model_code,
                        extra_compile_args=extra_compile_args,
                        model_name=self.model_name, verbose=verbose)
        try:
            with utils.suppress_stdout_stderr(False):
                self.fit = self.sm.sampling(
                                    data=self.data, iter=100,
                                    chains=1, init=0)
        except Exception as e:
            print('Error occurred during a sampling check for compiled model.')
            print('Could not sample for ' + model_code)
            raise e

        self.keys = self.fit.unconstrained_param_names()
        self.zlen = len(self.keys)

    def get_constrained_param_shapes(self):
        """ A function to get output names and shapes

        Returns:
            dict:
                A dictionary containing the shapes of all the parameters that
                are return my the PyStan's sampling function. This involves
                transformed parameters and generated quantities. This is useful
                to transform the unconstrained results from ADVI and vistan to
                constrained format similar to PyStan's StanModelFit4.extract()
        """
        results = self.fit.extract()
        N = self.get_n_samples(results)
        constrained_param_shapes = collections.OrderedDict()

        for k, v in results.items():
            if k == "lp__":
                continue
            assert(N == v.shape[0])
            constrained_param_shapes[k] = v.shape[1:]
        return constrained_param_shapes

    def z0(self):
        """
        Returns
        ----------
        np.ndarray
            a random initialization used for optimization
        """
        return np.random.randn(self.zlen)

    def sampling(self, verbose=False, **kwargs):
        """
        Args
        ----------
        verbose : bool
            If True, will print the suppressed print statements.
        kwargs : dict
            keyword arguments passed to PyStan's StanModel.sampling

        Returns
        ----------
        np.ndarray
            An array of unconstrained latent variables of the shape
            (num_samples, self.zlen)
        """
        try:
            assert kwargs.get('iter', 2) >= 2
            with utils.suppress_stdout_stderr(verbose):
                logging.getLogger("pystan").propagate = verbose
                self.fit = self.sm.sampling(data=self.data, **kwargs)
                logging.getLogger("pystan").propagate = False
            rez = self.unconstrain(self.fit.extract())
        except Exception as e:
            print("Error during sampling from the model.")
            raise e
        return rez

    def argmax(self, with_rez=False, method='BFGS', **kwargs):
        """
        Args
        ----------
            with_rez :  Boolean
                If True, returns the entire results object along
                with z_map.
            method :    String
                One of suggested_solvers = ['CG','BFGS','L-BFGS-B',
                                                'TNC','SLSQP'].
            kwargs:  dict
                Arguments for scipy.optimize.minimize. See
                https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html
                for more details
        Returns
        ----------
            np.ndarray, dict (optional)
            parameter at Maximum-a-posteriori, result object from
            scipy.optimize.minimize
        """
        suggested_solvers = ['CG', 'BFGS', 'L-BFGS-B', 'TNC', 'SLSQP']
        if method not in suggested_solvers:
            warnings.warn(
                        "Solver" + str(method)
                        + " passed to argmax not in suggested list"
                        + str(suggested_solvers)
                        + '\nsee list at:'
                        + 'https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html')
        z = self.z0()
        obj = autograd.value_and_grad(lambda z: -self.logp(z), 0)
        kwargs['disp'] = False
        rez = scipy.optimize.minimize(
                            obj, z,
                            method=method, jac=True, options=kwargs)
        z = rez.x
        if with_rez:
            return rez.x, rez
        else:
            return rez.x

    def advi(self, algorithm='fullrank', verbose=False, **kwargs):
        """
            A function to run Stan's variational Bayes method (ADVI)
            Args
            ----------
                algorithm:  "fullrank" or "meanfield\
                                    "
                verbsoe:    Boolean
                            If True, prints the optimization messages from ADVI
                kwargs:     arguments for vb see
                            https://pystan.readthedocs.io/en/latest/api.html#pystan.StanModel.vb
                            for more details
            Returns
            ----------
                samples:    samples from the final posterior
        """
        try:
            with utils.suppress_stdout_stderr(verbose):
                rez = self.sm.vb(
                            data=self.data,
                            algorithm=algorithm, **kwargs)

        except Exception as e:
            print('Error during ADVI...')
            raise e

        samples = np.array(rez["sampler_params"])[:-1, :].T

        return self.unconstrain(self.constrained_array_to_dict(samples))

    def mf(self, verbose=False, **kwargs):
        """
            A function to run Stan's variational Bayes method
            (ADVI) -- meanfield
        """
        return self.advi(algorithm='meanfield', verbose=verbose, **kwargs)

    def constrained_array_to_dict(self, samples):
        """ A function to convert the constrained parameter
            output to dictionary format compatible with
            PyStan's StanModelFit4.extract() format.

        Args:
            samples (np.ndarray): Constrained params

        Returns:
            dict:
                A dictionary that consists of the parameters aranged in
                the same format as PyStan's StanModelFit4.extract()
        """
        assert samples.ndim == 2
        N = samples.shape[0]

        constrained_param_shapes = self.get_constrained_param_shapes()
        params = collections.OrderedDict()
        idx = 0
        for param_name, param_shape in constrained_param_shapes.items():
            nparam = int(np.prod(param_shape))
            params[param_name] = np.reshape(
                                        samples[:, idx:idx+nparam],
                                        (N, *param_shape), order="F")
            idx += nparam
        assert idx == samples.shape[1]
        return params

    def constrain(self, z):
        """
            A function to constrain the parameters from unconstrained
            to original constrained range
            as defined in the model.
            Args
            ----------
                z:  samples in array np.ndarray format
            Returns
            ----------
                np.ndarray:
                    unconstrained samples np.ndarray format
        """

        assert z.ndim == 2
        assert z.shape[-1] == self.zlen
        return np.array([
                        self.fit.constrain_pars(np.asarray(z_, order='C'))
                        for z_ in z])

    def get_n_samples(self, z):
        """
            A function to get the number of samples.
            Args
            ----------
                z:  samples in array np.ndarray or dictionary format
            Returns
            ----------
                N: number of samples
        """

        if not isinstance(z, dict):
            assert z.ndim == 2
            return z.shape[0]
        else:
            return list(z.values())[0].shape[0]

    def unconstrain(self, z):
        """
            A function to unconstrain the parameters from original
            constrained range to unconstrained
            domain. Uses the model definition from Stan to transform.
            Args
            ----------
                z:  samples in array np.ndarray or dictionary format
            Returns
            ----------
                samples: unconstrained samples np.ndarray format
        """
        assert isinstance(z, dict)
        N = self.get_n_samples(z)

        return np.array([
                        self.fit.unconstrain_pars({
                                        k: v[n] for k, v in z.items()})
                        for n in range(N)])

    def log_prob(self, z):
        """
            A simple function to get batched sample evaluation for models.
            Args
            ----------
                z:  unconstrained samples in array np.ndarray
            Returns
            ----------
                log p: model log p(z,x) evaluation at unconstrained z samples.
                Adjust transform in used under the hood.
        """
        orig_samples_shape = z.shape
        assert(orig_samples_shape[-1] == self.zlen)
        lp = np.array([self.logp(z_) for z_ in z.reshape(-1, self.zlen)])

        return lp.reshape(orig_samples_shape[:-1])

    @autograd.primitive
    def logp(self, z):
        """
            A simple function to get log prob on a single sample
            Args
            ----------
                z (np.ndarray):
                single unconstrained sample
            Returns
            ----------
                log p(z,x) evaluation at unconstrained z
        """
        assert(z.ndim == 1)
        assert(len(z) == self.zlen)

        try:
            # try to evaluate logp with stan. if you fail, just return nan
            return self.fit.log_prob(z, adjust_transform=True)
        except Exception as e:
            return np.nan

    def glogp(self, z):
        """
            A simple function to get grad(log prob) on a single sample

            Args
            ----------
                z (np.ndarray):
                single unconstrained sample
            Returns
            ----------
                grad log p(z,x) evaluation at unconstrained z
        """
        assert(len(z) == self.zlen)
        rez_from_stan = self.fit.grad_log_prob(z, adjust_transform=True)
        return rez_from_stan.reshape(z.shape)

    def vjpmaker(argnum, rez, stuff, args):

        """
            A simple function to covert the grad to vjp
        """
        obj = stuff[0]
        z = stuff[1]
        if np.isnan(rez):
            return lambda gg: 0*z  # special gradient for nan case
        else:
            return lambda gg: gg*Model.glogp(obj, z)

###########################################################################
#  magic that makes autograd treat stan model logp as primitive function
###########################################################################


autograd.extend.defvjp_argnum(Model.logp, Model.vjpmaker)



import os
import pickle
import time
import functools
import operator
import collections
import warnings
import traceback
import hashlib
import re
import scipy
import autograd
import autograd.numpy as np
import autograd.numpy.random as npr
import autograd.misc.optimizers as optim
import tqdm.auto
import tqdm


def warning_on_one_line(message, category, filename, lineno, file=None,
                        line=None):
    return '%s:%s\n' % (category.__name__, message)


warnings.formatwarning = warning_on_one_line

###############################################################################
# Generic Inference Utilities
###############################################################################


def update_hparams(model, hparams):
    hparams['latent_dim'] = model.zlen

    # update_hparams_method(hparams)
    if hparams['fix_sample_budget']:
        hparams['num_copies_training'] = (
                                    hparams['per_iter_sample_budget']
                                    // hparams['M_iw_train'])
    else:
        hparams['num_copies_training'] = hparams['per_iter_sample_budget']
    # If advi is not used, then we follow the the step-size scaling
    # scheme of Agrawal et al. 2020
    if hparams['comprehensive_step_search_scaling']:
        hparams['step_size'] = hparams['step_size_base'] /\
                                    (hparams['step_size_scale']
                                        ** hparams['step_size_exp'])
        hparams['step_size'] = hparams['step_size']/hparams['latent_dim']


def get_recipe_hparams(method, hparams):
    if method == 'advi':
        hparams.update({
                'advi_use': True,
                'advi_convergence_threshold': 0.001,
                'advi_step_size': 1,
                'advi_adapt_step_size': True,
                'advi_adapt_step_size_range': [100, 10, 1, 0.1, 0.01],
                'advi_adapt_step_size_verbose': False,
                'advi_adapt_step_size_num_iters': 200,
                'advi_callback_iteration': 2,

                'vi_family': "gaussian",
                'grad_estimator': "closed-form-entropy",
                'optimizer': 'advi',
                'M_iw_train': 1,

                'LI': False,
                'per_iter_sample_budget': 100,
                'max_iters': 100,

                'evaluation_fn': 'ELBO-cfe',

                'fix_sample_budget': False,
                "comprehensive_step_search_scaling": False,


            })

    elif method == 'fullrank':
        hparams.update({
                'advi_use': False,
                'vi_family': "gaussian",

                "comprehensive_step_search_scaling": False,
                'step_size': 0.01,
                'max_iters': 100,
                'optimizer': 'adam',
                'M_iw_train': 1,
                'grad_estimator': "DReG",
                'per_iter_sample_budget': 100,

                'LI': False,

                'evaluation_fn': "IWELBO",

                'fix_sample_budget': False,


            })

    elif method == 'meanfield':
        hparams.update({

                'advi_use': False,
                'vi_family': "diagonal",

                "comprehensive_step_search_scaling": False,
                'step_size': 0.01,
                'max_iters': 100,
                'optimizer': 'adam',
                'M_iw_train': 1,
                'grad_estimator': "DReG",
                'per_iter_sample_budget': 100,

                'LI': False,
                'evaluation_fn': "IWELBO",
                'fix_sample_budget': False,

            })

    elif method == 'flows':
        hparams.update({

                'advi_use': False,
                'vi_family': "rnvp",

                "comprehensive_step_search_scaling": False,
                'step_size': 0.01,
                'max_iters': 100,
                'optimizer': 'adam',
                'M_iw_train': 1,
                'grad_estimator': "DReG",
                'per_iter_sample_budget': 100,

                'LI': False,

                'evaluation_fn': "IWELBO",

                'rnvp_num_transformations':  10,
                'rnvp_num_hidden_units':  16,
                'rnvp_num_hidden_layers':  2,
                'rnvp_params_init_scale':  0.01,

                'fix_sample_budget': False,
            })

    elif method == 'method 0':
        hparams.update({

                'advi_use': False,
                'vi_family': "gaussian",

                "comprehensive_step_search_scaling": True,
                'step_size_exp': 0,
                'step_size_base': 0.01,
                'step_size_scale': 4.0,
                'max_iters': 100,
                'optimizer': 'adam',
                'M_iw_train': 1,
                'grad_estimator': "closed-form-entropy",
                'per_iter_sample_budget': 100,

                'LI': False,

                'evaluation_fn': "IWELBO",

                'fix_sample_budget': True,

            })

    elif method == 'method 1':
        hparams.update({

                'advi_use': False,
                'vi_family': "gaussian",

                "comprehensive_step_search_scaling": True,
                'step_size_exp': 0,
                'step_size_base': 0.01,
                'step_size_scale': 4.0,
                'max_iters': 100,
                'optimizer': 'adam',
                'M_iw_train': 1,
                'grad_estimator': "STL",
                'per_iter_sample_budget': 100,

                'LI': False,

                'evaluation_fn': "IWELBO",

                'fix_sample_budget': True,

            })

    elif method == 'method 2':
        hparams.update({

                'advi_use': False,
                'vi_family': "gaussian",

                "comprehensive_step_search_scaling": True,
                'step_size_exp': 0,
                'step_size_base': 0.01,
                'step_size_scale': 4.0,
                'max_iters': 100,
                'optimizer': 'adam',
                'M_iw_train': 1,
                'grad_estimator': "STL",
                'per_iter_sample_budget': 100,

                'LI': True,
                "LI_max_iters": 2000,
                "LI_epsilon": 1e-6,

                'evaluation_fn': "IWELBO",

                'fix_sample_budget': True,

            })

    elif method == 'method 3a':
        warnings.warn("Use IW-sampling at inference to follow Method\
                                3a from the paper. Set M_iw_sample = 10")
        hparams.update({

                'advi_use': False,
                'vi_family': "gaussian",

                "comprehensive_step_search_scaling": True,
                'step_size_exp': 0,
                'step_size_base': 0.01,
                'step_size_scale': 4.0,
                'max_iters': 100,
                'optimizer': 'adam',
                'M_iw_train': 1,
                'grad_estimator': "STL",
                'per_iter_sample_budget': 100,
                'LI': False,

                'evaluation_fn': "IWELBO",

                'fix_sample_budget': True,

            })

    elif method == 'method 3b':
        hparams.update({

                'advi_use': False,
                'vi_family': "gaussian",

                "comprehensive_step_search_scaling": True,
                'step_size_exp': 0,
                'step_size_base': 0.01,
                'step_size_scale': 4.0,
                'max_iters': 100,
                'optimizer': 'adam',
                'M_iw_train': 10,
                'grad_estimator': "DReG",
                'per_iter_sample_budget': 100,

                'LI': False,

                'evaluation_fn': "IWELBO",

                'fix_sample_budget': True,

            })

    elif method == 'method 4a':
        hparams.update({

                'advi_use': False,
                'vi_family': "rnvp",

                "comprehensive_step_search_scaling": True,
                'step_size_exp': 0,
                'step_size_base': 0.01,
                'step_size_scale': 4.0,
                'max_iters': 100,
                'optimizer': 'adam',
                'M_iw_train': 1,
                'grad_estimator': "Total-gradient",
                'per_iter_sample_budget': 100,

                'LI': False,

                'evaluation_fn': "IWELBO",

                'rnvp_num_transformations':  10,
                'rnvp_num_hidden_units':  32,
                'rnvp_num_hidden_layers':  2,
                'rnvp_params_init_scale':  0.01,

                'fix_sample_budget': True,
            })

    elif method == 'method 4b':
        hparams.update({

                'advi_use': False,
                'vi_family': "rnvp",

                "comprehensive_step_search_scaling": True,
                'step_size_exp': 0,
                'step_size_base': 0.01,
                'step_size_scale': 4.0,
                'max_iters': 100,
                'optimizer': 'adam',
                'M_iw_train': 1,
                'grad_estimator': "DReG",
                'per_iter_sample_budget': 100,

                'LI': False,

                'evaluation_fn': "IWELBO",

                'rnvp_num_transformations':  10,
                'rnvp_num_hidden_units':  32,
                'rnvp_num_hidden_layers':  2,
                'rnvp_params_init_scale':  0.01,


                'fix_sample_budget': True,

            })

    elif method == 'method 4c':
        warnings.warn("Use IW-sampling at inference to follow \
                        Method 4c from the paper. Set M_iw_sample = 10")
        hparams.update({

                'advi_use': False,
                'vi_family': "rnvp",

                "comprehensive_step_search_scaling": True,
                'step_size_exp': 0,
                'step_size_base': 0.01,
                'step_size_scale': 4.0,
                'max_iters': 100,
                'optimizer': 'adam',
                'M_iw_train': 1,
                'grad_estimator': "DReG",
                'per_iter_sample_budget': 100,

                'LI': False,

                'evaluation_fn': "IWELBO",

                'rnvp_num_transformations':  10,
                'rnvp_num_hidden_units':  32,
                'rnvp_num_hidden_layers':  2,
                'rnvp_params_init_scale':  0.01,

                'fix_sample_budget': True,

            })

    elif method == 'method 4d':
        hparams.update({

                'advi_use': False,
                'vi_family': "rnvp",

                "comprehensive_step_search_scaling": True,
                'step_size_exp': 0,
                'step_size_base': 0.01,
                'step_size_scale': 4.0,
                'max_iters': 100,
                'optimizer': 'adam',
                'M_iw_train': 10,
                'grad_estimator': "DReG",
                'per_iter_sample_budget': 100,

                'LI': False,

                'evaluation_fn': "IWELBO",

                'rnvp_num_transformations':  10,
                'rnvp_num_hidden_units':  32,
                'rnvp_num_hidden_layers':  2,
                'rnvp_params_init_scale':  0.01,

                'fix_sample_budget': True,
            })

    else:
        raise NotImplementedError(f"""Method not unsupported.
            Expected one of
            ['advi', 'fullrank', 'meanfield', 'flows',
            'method 0', 'method 1', 'method 2', 'method 3b', 'method 3a',
            'method 4a', 'method 4b', 'method 4c', method 4d'],
             but got '{method}'
            """)


S = np.log(np.exp(1) - 1)


def good_iter(i):
    a = 10 ** np.floor(np.log10(i*1.0))
    return (i % a) == 0


def log_add_exp(x1, x2):
    return np.maximum(x1, x2) + np.log1p(np.exp(-np.abs(x1-x2)))


def log_sub_exp(x1, x2):
    assert (all(x1 > x2))
    return x1 + np.log1p(-np.exp(x2-x1))


def pos_diag(x):
    assert x.ndim == 1
    return log_add_exp(x, 0)


def pos_tril(x):
    assert x.ndim == 2
    return np.tril(x, -1) + np.diag(pos_diag(np.diag(x)))


def inv_pos_diag(x):
    assert x.ndim == 1
    return log_sub_exp(x, 0)


def inv_pos_tril(x):
    assert x.ndim == 2
    return np.tril(x, -1) + np.diag(inv_pos_diag(np.diag(x)))


def mul_iterable(x):
    return functools.reduce(operator.mul, x, 1)


###############################################################################
# Laplace's Initialization Utilities
###############################################################################


def Hessian_finite_differences(z, grad_f, ε):

    D = len(z)
    H = np.zeros((D, D))

    for d in range(D):
        z_pos = z*1.0
        z_pos[d] += ε
        z_neg = z*1.0
        z_neg[d] -= ε

        H[:, d] = (grad_f(z_pos) - grad_f(z_neg))/(2*ε)

    return H


def get_laplaces_init_params(log_p, z_len, num_epochs, ε=1e-4):
    """ A function to generate Laplaces approximation

    Args:
        log_p ([function]): [An autograd differentiable function]
        z_len ([int]): # of latent dimensions
        num_epochs ([int]): # of iterations

    Returns:
        [tuple]: [MAP estimate and the cholesky factor of \
                        inverse of negative Hessian at MAP estimate]
    """

    z_0 = npr.rand(z_len)
    # using minimize to maximize
    val_and_grad = autograd.value_and_grad(lambda z: -log_p(z))

    rez = scipy.optimize.minimize(
                                    val_and_grad, z_0,
                                    method='BFGS', jac=True,
                                    options={
                                            'maxiter': num_epochs,
                                            'disp': True})

    mu = rez.x
    H = Hessian_finite_differences(
                                    z=mu,
                                    grad_f=autograd.grad(lambda z: log_p(z)),
                                    ε=ε)
    try:
        neg_H_inv = np.linalg.inv(-H)
        L = np.linalg.cholesky(neg_H_inv)  # -H_inv = inv(-H)
        # modification to adjust for pos_tril function
        L = inv_pos_tril(L)
    except Exception as e:
        print('Using noisy unit covariance...')
        L = np.tril(0.1*npr.randn(z_len, z_len), -1) + np.eye(z_len)
        # modification to adjust for pos_tril function
        L = inv_pos_tril(L)

    return mu, L


def standardize_code(text):
    # text = re.sub('//.*?\n|/\*.*?\*/', '', text, flags=re.S)
    text = re.sub(r'//.*?\n|/\*.*?\*/', '', text, flags=re.S)
    pat = re.compile(r'\s+')
    text = pat.sub('', text)
    return text


def get_cache_fname(model_name, model_code):
    model_code_stripped = standardize_code(model_code)
    code_hash = hashlib.md5(model_code_stripped.encode('ascii')).hexdigest()
    if model_name is None:
        return code_hash
    else:
        return model_name + "_" + code_hash


def get_laplaces_init(log_p, z_len, num_epochs, ε, model_name, model_code):
    if check_laplaces_init_saved(model_name, model_code):
        return load_saved_laplaces_init(model_name, model_code)
    else:
        LI_params = get_laplaces_init_params(log_p, z_len, num_epochs, ε)
        save_laplaces_init(LI_params, model_name, model_code)
        return LI_params


def laplaces_init_dir():
    return 'data/laplaces_init/'


def laplaces_init_file(dir_name, model_name, model_code):
    return dir_name + get_cache_fname(model_name, model_code)\
                                                + "_LI_params.pkl"


def check_laplaces_init_saved(model_name, model_code):
    """
    Check if stored Laplaces Initialization parameters  are available.
    """
    file_name = laplaces_init_file(laplaces_init_dir(), model_name, model_code)
    if os.path.exists(file_name):
        return True
    return False


def load_saved_laplaces_init(model_name, model_code):
    file_name = laplaces_init_file(laplaces_init_dir(), model_name, model_code)
    if not os.path.exists(file_name):
        raise ValueError
    return open_pickled_files(file_name)


def save_laplaces_init(params,  model_name, model_code):
    dir_name = laplaces_init_dir()
    if not os.path.exists(dir_name):
        os.makedirs(dir_name)
    dump_pickled_files(
                        filename=laplaces_init_file(
                                                    dir_name,
                                                    model_name, model_code),
                        objects=params)


############################################################################
# Optimization Utilities
############################################################################


def advi_asserts(hparams):

    assert(hparams['advi_use'] == 1)
    assert(hparams['vi_family'] == 'gaussian')
    assert(hparams['M_iw_train'] == 1)
    assert(hparams['LI'] == 0)
    assert(hparams['grad_estimator'] == 'closed-form-entropy')
    assert(hparams['optimizer'] == 'advi')


def get_callback_arg_dict(hparams):

    if hparams['advi_use'] is True:

        buffer_len = np.int(max(
                                0.01*hparams['max_iters'] /
                                hparams['advi_callback_iteration'], 2))
        delta_results = collections.deque(maxlen=buffer_len)

        return {"delta_results": delta_results, "hparams": hparams}

    else:

        return {}


def run_optimization(
                        objective_grad, init_params, step_size,
                        num_epochs, callback, optimizer):

    optimized_params = optimizer(
                                grad=objective_grad,
                                x0=init_params,
                                step_size=step_size,
                                num_iters=num_epochs,
                                callback=callback)

    return optimized_params


def get_adapted_step_size(
                            objective_grad, eval_function, init_params,
                            optimizer, num_epochs, hparams):

    """
    Implements the adaptive step-size scheme from the PyStan version of ADVI.
    """

    init_elbo = eval_function(params=init_params)
    best_elbo = -1.0*np.inf
    best_step_size = 0

    try:
        print(f" Initial elbo: {init_elbo}")
        for i, step_size in enumerate(hparams['advi_adapt_step_size_range']):

            results = []
            print(f" Checking the step_size: {step_size}")
            try:
                optimized_params = run_optimization(
                        objective_grad=objective_grad,
                        init_params=init_params,
                        step_size=step_size,
                        num_epochs=hparams['advi_adapt_step_size_num_iters'],
                        callback=None,
                        optimizer=advi_optimizer)

            except Exception:
                print(f"Error occured during the optimization\
                                with step-size {step_size}...")
                print(traceback.print_exc())
                print(f"Using initial parameters instead for {step_size}...")
                optimized_params = init_params

            candidate_elbo = eval_function(optimized_params)

            if np.isnan(candidate_elbo):

                candidate_elbo = -1.0*np.inf

            if (candidate_elbo < best_elbo) & \
               (best_elbo > init_elbo):

                assert(best_step_size != 0)

                print("Best step_size found, best step_size:", best_step_size)
                print("Best step_size found, best elbo:", best_elbo)

                return best_step_size

            else:

                if ((i+1) < len(hparams['advi_adapt_step_size_range'])):

                    best_elbo = candidate_elbo
                    best_step_size = step_size

                else:

                    if candidate_elbo > init_elbo:
                        print("Best step_size found, \
                            best step_size:", best_step_size)
                        print("Best step_size found, best elbo:", best_elbo)
                        return best_step_size
                    else:
                        raise ValueError("ELBO value diverged \
                                for all step_sizes. Update step_size range")
    except Exception as e:
        print("Error occurred during when adapting step_size for ADVI")
        raise e


def optimization_handler(
                        objective_grad, eval_function,
                        init_params, optimizer,
                        num_epochs, step_size, callback, hparams, **kwargs):

    if(hparams['advi_adapt_step_size']) & (hparams['advi_use']):
        with suppress_stdout_stderr(hparams['advi_adapt_step_size_verbose']):
            step_size = get_adapted_step_size(
                                objective_grad=objective_grad,
                                eval_function=eval_function,
                                init_params=init_params,
                                optimizer=optimizer,
                                num_epochs=num_epochs,
                                hparams=hparams)

    results = []
    t0 = time.time()
    optimized_params = run_optimization(
                            objective_grad=objective_grad,
                            init_params=init_params,
                            num_epochs=num_epochs,
                            step_size=step_size,
                            callback=functools.partial(
                                            callback,
                                            results=results,
                                            **get_callback_arg_dict(hparams)),
                            optimizer=optimizer)

    tn = time.time() - t0
    return (results, tn, len(results)), optimized_params


def callback(params, t, g, results, model, eval_function):

    results.append(eval_function(params))

    if good_iter(t+1):
        if np.isnan(results[-1]):
            print("exiting optimization because nan encountered.")
            return "exit"
    return None


def relative_difference(curr, prev):
    return np.abs((curr-prev)/prev)


def advi_callback(
                params, t, g, results, delta_results, model,
                eval_function, hparams):
    results.append(eval_function(params))

    if (t+1) % hparams['advi_callback_iteration'] == 0:
        tqdm.tqdm.write(f"Iteration {t+1}")
        tqdm.tqdm.write(f"ELBO, running mean  :{np.nanmean(results)}")
        tqdm.tqdm.write(f"ELBO, current value :{results[-1]}")

        if len(results) > hparams['advi_callback_iteration']:
            previous_elbo = results[-(hparams['advi_callback_iteration']+1)]
        else:
            previous_elbo = 0.0

        current_elbo = results[-1]
        delta_results.append(relative_difference(previous_elbo, current_elbo))
        delta_elbo_mean = np.nanmean(delta_results)
        delta_elbo_median = np.nanmedian(delta_results)

        tqdm.tqdm.write(f"Tolerance Δ, mean {delta_elbo_mean}")
        tqdm.tqdm.write(f"Tolerance Δ, median {delta_elbo_median}")
        if((delta_elbo_median <= hparams['advi_convergence_threshold']) |
                (delta_elbo_mean <= hparams['advi_convergence_threshold'])):
            tqdm.tqdm.write("Converged according to ADVI \
                metrics for Median/Mean")
            return "exit"
    return None


def adam(grad, x0, callback=None, num_iters=100,
         step_size=0.001, b1=0.9, b2=0.999, eps=10**-8):
    """Adam as described in http://arxiv.org/pdf/1412.6980.pdf.
    It's basically RMSprop with momentum and some correction terms."""
    x, unflatten = autograd.misc.flatten(x0)
    m = np.zeros(len(x))
    v = np.zeros(len(x))
    for i in tqdm.auto.tqdm(range(num_iters), desc="Optimizing"):
        g = autograd.misc.flatten(grad(unflatten(x), i))[0]
        if callback:
            flag = callback(unflatten(x), i, unflatten(g))
            if flag == "exit":
                return unflatten(x)

        m = (1 - b1) * g + b1 * m  # First  moment estimate.
        v = (1 - b2) * (g**2) + b2 * v  # Second moment estimate.
        mhat = m / (1 - b1**(i + 1))    # Bias correction.
        vhat = v / (1 - b2**(i + 1))
        x = x - step_size*mhat/(np.sqrt(vhat) + eps)
    return unflatten(x)


def advi_optimizer(
                    grad, x0, callback, num_iters, step_size,
                    epsilon=1e-16, tau=1, alpha=0.1):

    """ADVI optimizer as described
     in https://dl.acm.org/doi/pdf/10.5555/3122009.3122023.
    """
    x, unflatten = autograd.misc.flatten(x0)
    s = np.zeros(len(x))
    for i in tqdm.auto.tqdm(range(num_iters), desc="Optimizing"):
        g = autograd.misc.flatten(grad(unflatten(x), i))[0]
        if callback:
            flag = callback(unflatten(x), i, unflatten(g))
            if flag == "exit":
                return unflatten(x)
        if i == 0:
            s = g**2
        else:
            s = alpha*(g**2) + (1-alpha)*s
        x = x - (step_size / np.sqrt(i+1.))*g / (tau + np.sqrt(s))
    return unflatten(x)


def get_step_size(hparams):

    if hparams['advi_use'] is True:
        return hparams['advi_step_size']
    return hparams['step_size']


def get_LI_params(hparams, model, code, model_name):
    # update the initial parameters to LI params
    assert "gaussian" in hparams['vi_family']
    try:
        with suppress_stdout_stderr(False):
            init_params = get_laplaces_init(
                                    log_p=model.log_prob,
                                    z_len=hparams['latent_dim'],
                                    num_epochs=hparams['LI_max_iters'],
                                    ε=hparams['LI_epsilon'],
                                    model_name=model_name,
                                    model_code=code)
        return init_params
    except Exception as e:
        print("Error occurred trying to generate\
             Laplace's Initialization parameters.")


def get_optimizer(hparams):

    if hparams['optimizer'] == "adam":
        return adam
    elif hparams['optimizer'] == "advi":
        return advi_optimizer
    elif hparams['optimizer'] == 'sgd':
        warnings.warn("Using autograd's SGD. \
            Will not exit optimization if NaN occurs.")
        return optim.sgd
    elif hparams['optimizer'] == 'rmsprop':
        warnings.warn("Using autograd's RMSprop. \
            Will not exit optimization if NaN occurs.")
        return optim.rmsprop
    else:
        raise NotImplementedError


def get_callback(hparams):

    if hparams['advi_use'] == 0:
        return callback
    else:
        return advi_callback


##############################################################################
# Generic Flow Utilities
##############################################################################


def coupling_layer_specifications(num_hidden_units, num_hidden_layers, z_len):
    """
    We specify the FNN based networks over here. A single network
    produce both s and t parts.
    Coupling Layer currently comprises of 2 transforms.
    """

    d_1 = np.int(z_len//2)
    d_2 = np.int(z_len - d_1)
    coupling_layer_sizes = []
    coupling_layer_sizes.append([d_1] +
                                num_hidden_layers*[num_hidden_units] + [2*d_2])
    coupling_layer_sizes.append([d_2] +
                                num_hidden_layers*[num_hidden_units] + [2*d_1])
    return coupling_layer_sizes


##############################################################################
# Generic NN Utilities
##############################################################################

def relu(x):
    return np.maximum(0, x)


def leakyrelu(x, slope=0.01):
    return np.maximum(0, x) + slope*np.minimum(0, x)


def tanh(x):
    return np.tanh(x)


def softmax_matrix(x):
    assert x.ndim <= 2
    y = x.T
    z = y - np.max(y, 0)
    z = np.exp(z)
    return (z/np.sum(z, 0)).T


############################################################################
# Generic file Utilities
############################################################################


def dump_pickled_files(filename, objects, protocol=None):
    with open(filename, "wb") as f:
        if protocol is None:
            pickle.dump(objects, f)
        else:
            pickle.dump(objects, f, protocol=protocol)


def open_pickled_files(filename, protocol=None):
    with open(filename, "rb") as f:
        if protocol is None:
            objects = pickle.load(f)
        else:
            objects = pickle.load(f, protocol=protocol)
        return objects


############################################################################
# Generic Print Utilities
############################################################################

def print_hparams(hparams):
    # hparams['step_size'] = hparams['step_size']/hparams['latent_dim']
    # using a fixed sample budget--we need to adjust the no. of
    # training samples based on M
    common_hparams = [
                        'num_copies_training', 'per_iter_sample_budget',
                        'latent_dim', 'evaluation_fn', 'grad_estimator',
                        'vi_family', 'max_iters', 'optimizer',
                        'M_iw_train', 'method']

    # print the common relevant hparams
    for k, i in hparams.items():
        if k in common_hparams:
            print(f"{k}:{i}")

    if hparams['method'] == 'advi':
        # print the advi relevant hparams
        for k, i in hparams.items():
            if ('advi' in k):
                print(f"{k}:{i}")
    elif hparams['method'] in ['gaussian', 'flows', 'meanfield']:
        # print the other relevant hparams
        for k, i in hparams.items():
            if (k.startswith("step")):
                print(f"{k}:{i}")
            if (k.startswith("rnvp")) & (hparams['method'] == 'flows'):
                print(f"{k}:{i}")
    elif hparams['method'] == 'custom':
        # print the remaining hparams for custom
        for k, i in hparams.items():
            if k not in common_hparams:
                print(f"{k}:{i}")

    else:
        raise ValueError("Method should be one of \
            ['custom', 'advi', 'gaussian', 'rnvp', 'meanfield']")


class suppress_stdout_stderr(object):
    '''
    A context manager for doing a "deep suppression" of stdout and stderr in
    Python, i.e. will suppress all print, even if the print originates in a
    compiled C/Fortran sub-function.
       This will not suppress raised exceptions, since exceptions are printed
    to stderr just before a script exits, and after the context manager has
    exited (at least, I think that is why it lets exceptions through).

    '''
    def __init__(self, verbose):
        self.verbose = verbose
        # Open a pair of null files
        self.null_fds = [os.open(os.devnull, os.O_RDWR) for x in range(2)]
        # Save the actual stdout (1) and stderr (2) file descriptors.
        self.save_fds = [os.dup(1), os.dup(2)]

    def __enter__(self):

        if not self.verbose:
            # Assign the null pointers to stdout and stderr.
            os.dup2(self.null_fds[0], 1)
            os.dup2(self.null_fds[1], 2)

    def __exit__(self, *_):

        # Re-assign the real stdout/stderr back to (1) and (2)
        if not self.verbose:

            os.dup2(self.save_fds[0], 1)
            os.dup2(self.save_fds[1], 2)
            # Close all file descriptors
            for fd in self.null_fds + self.save_fds:
                os.close(fd)



import autograd.numpy as np
import autograd.numpy.random as npr
import functools
import collections
import warnings
import vistan.utilities as utils

warnings.formatwarning = utils.warning_on_one_line


class Dist():
    """
        A template class for distributions.
    """

    def __init__(self, zlen):

        self.zlen = zlen

    def initial_params(self):

        raise NotImplementedError

    def transform_params(self, params):

        raise NotImplementedError

    def get_params(self, params):

        return self.transform_params(params)

    def sample(self, params, sample_shape):

        raise NotImplementedError

    def log_prob(self, params, samples):

        raise NotImplementedError


S = np.log(np.exp(1)-1)


class Gaussian(Dist):

    def initial_params(self):

        return [np.zeros(self.zlen,).astype(float),
                S*np.eye(self.zlen).astype(float)]

    def transform_params(self, params):

        return params[0], utils.pos_tril(params[1])

    def sample(self, params, sample_shape=()):

        if isinstance(sample_shape, int):
            sample_shape = (sample_shape,)

        mu, sig = self.get_params(params)
        shape = sample_shape+(self.zlen,)
        ε = npr.randn(*shape)
        samples = mu + np.dot(ε, sig.T)

        return samples

    def log_prob(self, params, samples):

        mu, sig = self.get_params(params)
        Λ = np.linalg.inv(sig).T
        M = (samples-mu)
        a = self.zlen*np.log(2*np.pi)
        b = 2*np.sum(np.log((np.diag(sig))))
        c = np.sum(np.matmul(M, np.matmul(Λ, Λ.T)) * M, -1)

        return -0.5*(a+b+c)

    def entropy(self, params):

        mu, sig = self.get_params(params)
        a = self.zlen*np.log(2*np.pi) + self.zlen
        b = 2*np.sum(np.log((np.diag(sig))))

        return 0.5*(a+b)


class Diagonal(Gaussian):
    """
        A class to implement Diagonal Gaussian distribution
    """
    def initial_params(self):
        return [np.zeros(self.zlen,).astype(float),
                S*np.ones(self.zlen,).astype(float)]

    def transform_params(self, params):
        return params[0], np.diag(utils.pos_diag(params[1]))


class Flows(Dist):

    def __init__(self, zlen):

        self.zlen = zlen
        self.base_dist = Gaussian(self.zlen)
        self.base_dist_params = self.base_dist.initial_params()

    def transform_params(self, params):

        return params

    def forward_transform(self, params, z):

        raise NotImplementedError

    def inverse_transform(self, params, x):

        raise NotImplementedError

    def sample(self, params, sample_shape):
        params = self.get_params(params)
        z_o = self.base_dist.sample(self.base_dist_params, sample_shape)
        samples, neg_log_det_J = self.forward_transform(params, z_o)

        return samples

    def log_prob(self, params, samples):

        params = self.get_params(params)

        z_o, neg_log_det_J = self.inverse_transform(params, samples)

        lq = self.base_dist.log_prob(self.base_dist_params, z_o)

        return lq + neg_log_det_J


def BinaryFlip(z, j):
    return np.concatenate(reversed(BinarySplit(z, j)), -1)


def ReverseBinaryFlip(z, j):
    return np.concatenate(reversed(ReverseBinarySplit(z, j)), -1)


def BinarySplit(z, j):

    D = z.shape[-1]
    d = D//2
    if D % 2 == 1:
        d += (np.int(j) % 2)

    return np.array_split(z, [d], -1)


def ReverseBinarySplit(z, j):
    return BinarySplit(z, np.int(j)+1)


class RealNVP(Flows):

    def __init__(
                self,
                num_transformations,
                num_hidden_units,
                num_hidden_layers,
                params_init_scale,
                zlen):

        self.num_transformations = num_transformations
        self.num_hidden_layers = num_hidden_layers
        self.num_hidden_units = num_hidden_units
        self.params_init_scale = params_init_scale

        super(RealNVP, self).__init__(zlen=zlen)

    def initial_params(self):

        def generate_net_st():

            coupling_layer_sizes = utils.coupling_layer_specifications(
                                self.num_hidden_units,
                                self.num_hidden_layers, self.zlen)

            init_st_params = []

            for layer_sizes in coupling_layer_sizes:

                init_st_params.append([(
                    self.params_init_scale * npr.randn(m, n),   # weight matrix
                    self.params_init_scale * npr.randn(n))      # bias vector
                    for m, n in
                    zip(layer_sizes[:-1], layer_sizes[1:])])
            return init_st_params

        st = [generate_net_st() for i in range(self.num_transformations)]

        return st

    def apply_net_st(self, params, inputs):

        inpW, inpb = params[0]

        inputs = utils.leakyrelu(np.dot(inputs, inpW) + inpb)

        for W, b in params[1:-1]:
            outputs = np.dot(inputs, W) + b
            inputs = utils.leakyrelu(outputs)

        outW, outb = params[-1]
        outputs = np.dot(inputs, outW) + outb

        assert(outputs.shape[:-1] == inputs.shape[:-1])
        assert(outputs.shape[-1] % 2 == 0)

        s, t = np.array_split(outputs, 2, -1)

        assert(s.shape == t.shape)

        return utils.tanh(s), t

    def forward_transform(self, params, z):

        neg_log_det_J = np.zeros(z.shape[:-1])

        st_list = params

        for i in range(self.num_transformations):
            for j in range(2):

                z_1, z_2 = BinarySplit(z, j)

                s, t = self.apply_net_st(params=st_list[i][j], inputs=z_1)
                z = np.concatenate([z_1, z_2*np.exp(s) + t], axis=-1)
                # Immediately update neg log jacobian with s outputs
                neg_log_det_J -= np.sum(s, axis=-1)

                z = BinaryFlip(z, j)

        assert(z.shape[:-1] == neg_log_det_J.shape)
        return z, neg_log_det_J

    def inverse_transform(self, params, x):

        neg_log_det_J = np.zeros(x.shape[:-1])

        st_list = params

        for i in reversed(range(self.num_transformations)):
            for j in reversed(range(2)):

                x = ReverseBinaryFlip(x, j)

                x_1, x_2 = BinarySplit(x, j)

                s, t = self.apply_net_st(
                                params=st_list[i][j],
                                inputs=x_1)
                # output should be of the shape of x_1

                x = np.concatenate(
                                    [x_1, (x_2 - t)*np.exp(-s)],
                                    axis=-1)
                neg_log_det_J -= np.sum(s, axis=-1)

        assert(x.shape[:-1] == neg_log_det_J.shape)

        return x, neg_log_det_J


def get_var_dist(hyper_params):

    if hyper_params['vi_family'] == "gaussian":

        var_dist = Gaussian(hyper_params['latent_dim'])

    elif hyper_params['vi_family'] == "diagonal":

        var_dist = Diagonal(hyper_params['latent_dim'])

    elif hyper_params['vi_family'] == "rnvp":

        var_dist = RealNVP(
                num_transformations=hyper_params['rnvp_num_transformations'],
                num_hidden_units=hyper_params['rnvp_num_hidden_units'],
                num_hidden_layers=hyper_params['rnvp_num_hidden_layers'],
                params_init_scale=hyper_params['rnvp_params_init_scale'],
                zlen=hyper_params['latent_dim'])
    else:
        raise NotImplementedError

    return var_dist


class Posterior():
    """
        A simple Posterior class.
    """
    def __init__(self, M_iw_sample, model, params, var_dist, results):
        """

        Args:
            M_iw_sample (int):
                # of importance samples to use
            model (vistan.interface.Model):
                An instance of vistan.interface.Model that was used
                during optimization
            params (tuple of np.ndarray):
                Final optimized params correponding variational approximation
            var_dist (vistan.vi_families.Dist):
                The final approximate distribution.
            results (tuple):
                optimization results
        """
        self.M_iw_sample = M_iw_sample
        self.params = params
        (
            self.optimization_trace,
            self.optimization_time,
            self.optimization_num_iters) = results
        self.model = model
        self.var_dist = var_dist

        self.log_p = self.model.log_prob
        self.log_q = self.var_dist.log_prob
        self.sample_q = self.var_dist.sample
        self.zlen = self.model.zlen

    def sample(
                self, num_samples, *, params=None,
                M_iw_sample=None, return_constrained=True):
        """
            A function to allows sampling from the posterior.

            Args
            ----------

                num_samples (int):
                    # of samples

                params :
                    if None, then the stored optimized variational parameters
                    are used to sample. Alternatively, one can provide
                    compatible parameters.

                M_iw_sample (int):
                    If None, then M_iw_sample is set to M_iw_train the
                    variational inference was optimized with. Otherwise, a
                    user can specify a different M_iw_sample. This scales the
                    sampling cost linearly.

                return_constrained (bool):
                    If True, returns the constrained parameters in the same
                    dictionary template as PyStan's .extract() method
                    (without 'lp__'.)
                    If False, then samples are returned unconstrained in
                    np.ndarray format such that z.shape =
                    (num_samples, self.zlen)
            Returns
            ----------

                np.ndarray or dict:
                    If return_constrained is True, then a dictionary is
                    the same template as PyStan's StanModelFit4.extract()
                    method. Else, returns the unconstrained samples
                    in the np.ndarray format such that
                    z.shape = (num_samples, self.zlen)

        """
        if params is None:
            params = self.params
        else:
            warnings.warn("Using params different then the optimized params.")

        if M_iw_sample is None:
            M_iw_sample = self.M_iw_sample
        elif M_iw_sample < self.M_iw_sample:
            warnings.warn("Specified M_iw_sample is less than the M_iw_train.")

        if M_iw_sample == 1:
            samples = self.sample_q(params, num_samples)
        else:
            shape = (num_samples, M_iw_sample)
            samples = self.sample_q(params, shape)

            lp = self.log_p(samples)
            lq = self.log_q(params, samples)
            lR = lp - lq
            sampler = np.vectorize(
                        lambda p:  np.random.multinomial(1, p),
                        signature="(n)->(n)")

            choices = sampler(utils.softmax_matrix(lR))
            samples = samples[np.where(choices)]

            assert samples.shape == (num_samples, self.zlen)
        if return_constrained is True:
            return self.constrained_array_to_dict(self.constrain(samples))
        else:
            return samples

    def log_prob(self, samples, params=None):
        """
            A function to get log_prob of the base q distribution. This is
            not necessarily same as the log_prob of the posterior.

            Args
            ----------

                samples (dict or np.ndarray):
                    samples as returned by the Posterior.sample function

                params :
                    if None, then the stored optimized variational parameters
                    are used to sample. Alternatively, one can provide
                    compatible parameters.

            Returns
            ----------

                np.ndarray or float:
                    The log_prob for base q distribution evaluated
                    in the unconstrained space. This is not always the
                    log_prob of the posterior distribution.
        """

        if params is None:
            params = self.params
        else:
            warnings.warn("Using params different then the optimized params.")

        warnings.warn("""
            log_prob function returns the log density of the
            unconstrained base q distribution used during optimization.
            If you trained with M_iw_train > 1, then this is different
            than the log_prob of the posterior.

            Please, see https://arxiv.org/pdf/1808.09034 for more details.
            """)

        if isinstance(samples, dict):
            samples = self.unconstrain(samples)
        return self.log_q(params, samples)

    def constrain(self, z):
        """A function to convert unconstrained latent variables to
            constained space. This also returns the other generated quantities
            as defined by the Stan code.

        Args:
            z (np.ndarray):
                Unconstrained space sampels.

        Returns:
            np.ndarray:
                Parameters in the constrained space along with other
                generated quantities.
        """
        return self.model.constrain(z)

    def constrained_array_to_dict(self, z):
        """A function to convert constrained space parameters to
            a nice dictionary format as used by PyStan's
            StanModelFit4.extract()

        Args:
            z (np.ndarray):
                np.ndarray as obtained form the call to 'constrain' function

        Returns:
            dict:
                See vistan.interface.Model.constrained_array_to_dict
        """
        assert z.ndim == 2
        return self.model.constrained_array_to_dict(z)

    def unconstrain(self, samples):
        """A function to convert the samples from constrained domain
         to constrained domain.

        Args:
            samples (dict):
                Dictionary of parameters. Dictionary can optionally contain
                other generated quantities or transformed parameters. Only the
                latent variables in the unconstrained space are returned.
                See Pystan's StanModelFit4.unconstrain for more information.

        Returns:
            np.ndarray: Unconstrained latent variables.
        """
        assert isinstance(samples, dict)
        return self.model.unconstrain(samples)


def get_posterior(model, var_dist, params, M_iw_sample, results):

    q = Posterior(
                    M_iw_sample=M_iw_sample,
                    model=model,
                    var_dist=var_dist,
                    params=params,
                    results=results)
    return q



import functools
import autograd.numpy as np
import autograd.scipy.special as autoscipy
import autograd
import vistan.utilities as utils


def objective_utils(
                    params, log_p, log_q, sample_q,
                    M_iw_train, num_copies_training):

    samples_shape = (num_copies_training, M_iw_train)

    params_stopped = autograd.core.getval(params)

    z = sample_q(params, samples_shape)

    lp = log_p(z)

    lq = log_q(params, z)

    lq_stopped = log_q(params_stopped, z)

    return z, lp, lq, lq_stopped


def ELBO_cf_entropy(
                    params, log_p, log_q, sample_q, entropy_q,
                    M_iw_train, num_copies_training):

    _, lp, _, _ = objective_utils(
                                params, log_p, log_q, sample_q,
                                M_iw_train, num_copies_training)

    return np.mean(lp) + entropy_q(params)


def IWELBO(params, log_p, log_q, sample_q, M_iw_train, num_copies_training):

    _, lp, lq, _ = objective_utils(
                                    params, log_p, log_q, sample_q,
                                    M_iw_train, num_copies_training)

    return np.mean(autoscipy.logsumexp(lp - lq, -1)) - np.log(M_iw_train)


def IWELBO_STL(
            params, log_p, log_q,
            sample_q, M_iw_train, num_copies_training):

    _, lp, _, lq_stopped = objective_utils(
                        params, log_p, log_q,
                        sample_q, M_iw_train, num_copies_training)

    lR = lp - lq_stopped

    return np.mean(np.sum(utils.softmax_matrix(lR)*lR, -1))


def IWELBO_DREG(
                params, log_p, log_q, sample_q,
                M_iw_train, num_copies_training):

    _, lp, _, lq_stopped = objective_utils(
                            params, log_p,
                            log_q, sample_q, M_iw_train, num_copies_training)

    lR = lp - lq_stopped

    return np.mean(np.sum((utils.softmax_matrix(lR)**2)*lR, -1))


def choose_objective_eval_fn(hyper_params):

    if hyper_params['grad_estimator'] == "Total-gradient":
        objective = IWELBO

    elif hyper_params['grad_estimator'] == "DReG":
        objective = IWELBO_DREG

    elif hyper_params['grad_estimator'] == "STL":
        objective = IWELBO_STL

    elif hyper_params['grad_estimator'] == "closed-form-entropy":
        assert (hyper_params['M_iw_train'] == 1)
        assert ("gaussian" in hyper_params['vi_family'])
        objective = ELBO_cf_entropy
    else:
        raise ValueError

    if hyper_params['evaluation_fn'] == "IWELBO":
        evaluation_fn = IWELBO

    elif hyper_params['evaluation_fn'] == "ELBO-cfe":
        evaluation_fn = ELBO_cf_entropy

    else:
        raise ValueError

    return objective, evaluation_fn


def modify_objective_eval_fn(
                    objective, evaluation_fn,
                    log_p, var_dist, hyper_params):

    m_objective = functools.partial(
                    objective,
                    log_p=log_p,
                    log_q=var_dist.log_prob,
                    sample_q=var_dist.sample,
                    M_iw_train=hyper_params['M_iw_train'],
                    num_copies_training=hyper_params['num_copies_training'])

    m_evaluation_fn = functools.partial(
                    evaluation_fn,
                    log_p=log_p,
                    log_q=var_dist.log_prob,
                    sample_q=var_dist.sample,
                    M_iw_train=hyper_params['M_iw_train'],
                    num_copies_training=hyper_params['num_copies_training'])

    if hyper_params['grad_estimator'] == "closed-form-entropy":
        m_objective = functools.partial(
                            m_objective,
                            entropy_q=var_dist.entropy)

    if hyper_params['evaluation_fn'] == "ELBO-cfe":
        m_evaluation_fn = functools.partial(
                            m_evaluation_fn,
                            entropy_q=var_dist.entropy)

    # Augment the objective definition to match
    # the Autograd's optimizer template
    return lambda params, t: -m_objective(params), m_evaluation_fn


def get_objective_eval_fn(log_p, var_dist, hyper_params):

    objective, evaluation_fn = choose_objective_eval_fn(hyper_params)

    return modify_objective_eval_fn(
                    objective, evaluation_fn, log_p,
                    var_dist, hyper_params)



import functools
import autograd
import autograd.numpy.random as npr

import vistan.vi_families as vi_families
import vistan.objectives as objectives
import vistan.interface as interface
import vistan.utilities as utils
import vistan.hyperparams as hyperparams


def recipe(method='meanfield', *, max_iters=100):
    """
        A function to easily run default variational methods--recipes.

        Args
        ----------
        method (string):
            One of ['advi', 'fullrank', 'meanfield', 'flows',
                    'method 0', 'method 1', 'method 2', 'method 3b',
                    'method 4a', 'method 4b', 'method 4d']
            'advi':
                Runs our implementation of ADVI
            'fullrank':
                Runs a version of fullrank Gaussian VI
            'meanfield':
                Runs a version of diagonal Gaussian VI
            'flows':
                Runs a version of flow-based VI using realNVP
            'method x':
                Runs a vi method as described in
                the paper https://arxiv.org/pdf/2006.10343.pdf
        max_iters (int):
            Changes the default number of iterations in recipes.
        Returns
        -------
        'inference' function with modified hyper-parameters locked into it.
    """
    default_hparams = hyperparams.default_hparams_dict.copy()
    utils.get_recipe_hparams(method, default_hparams)
    default_hparams["max_iters"] = max_iters
    return functools.partial(inference, hparams=default_hparams)


def algorithm(**kwargs):
    """
        A function to easily run custom variational methods.

        Args
        ----------

        kwargs (dict):

            Can contain the following attributes.
            ---------------
            ADVI hyperparams
            ---------------

            advi_use (bool):
                If true, alerts the internal functions that you are running
                ADVI as reported in
                paper https://www.jmlr.org/papers/volume18/16-107/16-107.pdf.
                You can custom use different functionalities of
                ADVI implementation; however, you would want to switch
                'advi_use' to False in that case.

            advi_convergence_threshold (float):
                The threshold parameters is used to diagnose convergence.
                If the mean or median of the relative tolerance window
                is below the threshold parameter, then
                the optimization is stopped.

            advi_step_size (float):
                parameter eta as defined in the original paper.

            advi_adapt_step_size (bool):
                Default is True. If True, then it uses the
                heuristic step-size scheme in the PyStan's ADVI implementation.

            advi_adapt_step_size_range (iterable):
                The range used in the original paper was [100,10,1,0.1,0.01]

            advi_adapt_step_size_num_iters (int):
                The number of iterations to use for each step-size candidate,
                ADVI greedily chooses the maximum step-size based on the final
                ELBO value after 'advi_adapt_step_size_num_iters' iterations.

            advi_callback_iterations (int):
                # of iterations after which ELBO is evaluated. This evaluation
                is then used to monitor convergence by monitoring the mean
                and the median of relative tolerance within a window of values.
                The length of this window (circular buffer)
                is heuristically determined.

            ---------------
            Optimization
            ---------------

            step_size (float):
                Default value is 0.1
                step_size = step_size_base/(step_size_scale**step_size_exp)
                To try custom step_sizes, use the step_size_base attribute

            comprehensive_step_search_scaling (bool):
                Default is False. If True, will scale the step-size using
                the heuristic used in the
                paper https://arxiv.org/pdf/2006.10343.pdf. If False, it
                will use the 'step_size' as is.

            step_size_base (float):
                Default value is 0.1.

            step_size_scale (float):
                Default value is 4.0

            step_size_exp (int):
                Default value is 0

            step_size_exp_range (iterable):
                Default value is [0,1,2,3]. This is currently not supported.
                For comprehensive step-search, we optimize for
                different step_sizes, where the different step_sizes are
                generated using step_size_exp_range.
                Comprehensive step-search is not supported right now.

            max_iters (int):
                # of optimization iterations.
                In case of ADVI, if we choose to adapt step-size, then
                this is only used for final adapted step_size.

            optimizer (string):
                One of ['adam', 'sgd', 'rmsprop', 'advi']
                'advi' is the custom step-size sequence
                proposed in the ADVI paper
                (see https://www.jmlr.org/papers/volume18/16-107/16-107.pdf
                page 12.)
                Currently, only 'advi' and 'adam' support early exit
                based on the callback function

            M_iw_train (int):
                # of importance weights during training to optimize
                the IW-ELBO. Setting M_iw_train=1 reverts to
                naive VI (optimizer ELBO).

            grad_estimator (string):
                One of ['DReG', 'STL', 'Total-gradient', 'closed-form-entropy']
                Total gradient refers to the regular IW-ELBO gradient.
                With M_iw_train = 1 this would be same as regular VI.
                'closed-form-entropy' works when families support
                closed-form-entropy calculation ("gaussian" or "diagonal".)

            per_iter_sample_budget (int):
                Fixes the # of samples used at each iteration to calculate
                the gradient and the evaluation_fn. If per_iter_sample_budget
                 = 100 and M_iw_train = 10, then we use 10 copies of
                 IW-ELBO gradient to optimize at each iteration.

            --------------------
            Evaluation Function
            --------------------

            evaluation_fn (string):
                One of ['IWELBO', 'ELBO-cfe']
                When using different gradient estimators like DREG and STL,
                the objective function is a surrogate function that has the
                same gradient as IW-ELBO but is different in value.
                evaluation_fn helps you monitor the correct objective.
                If set to IW-ELBO, it uses the same M as M_iw_train
                If set to ELBO-cfe, it monitors the ELBO using the
                closed-form entropy function
            -----------
            VI Families
            -----------

            vi_family (string):
                One of ["rnvp", "gaussian", "diagonal"]

            -------
            RealNVP
            -------

            rnvp_num_transformations (int):
                # of coupling transforms; each transforms includes
                two transitions

            rnvp_num_hidden_layers (int):
                # of hidden layers within s-t block

            rnvp_num_hidden_units (int):
                # of units in each hidden layer

            params_init_scale (int):
                scaling to help initializes to standard normal approximately

            ----------------------------
            LI: Laplace's Initialization
            ----------------------------

            LI (bool):
                Default is False. If True, initializes the
                Gaussian parameters to Laplace's approximation.
                Do not use LI with real-NVP

            LI_max_iters (int):
                Default is 2000. # of iterations for MAP estimate
                used by scipy.optimize.minimize

            LI_epsilon (float):
                Default is 1e-6. A small positive used to calculate
                Hessian at MAP estimate using finite differences.

            --------------------
            Other flexibilities
            --------------------

            fix_sample_budget (bool):
                Default is True. If False, will increase the
                per_iter_sample budget by M_iw_train times. This
                sets the 'num_copies_training' parameter to be same
                as 'per_iter_sample_budget' which combined with
                'M_iw_train' decides the final number of samples at
                each iteration (num_copies_training, M_iw_train).



        Returns
        -------

        'inference' function with modified hyper-parameters locked into it.
    """

    hparams = hyperparams.default_hparams_dict.copy()

    for k in kwargs.keys():
        if k not in hparams.keys():
            raise KeyError(f"""{k} is not an expected hyperparam.
                Hyper-param should be one of {list(hparams.keys())}""")

    # update the arguments to override the default arguments
    hparams.update(kwargs)

    return functools.partial(inference, hparams=hparams)


def inference(
            code, data, hparams, *, model_name="default_model_name",
            verbose=False, print_hparams=False):
    """
        A function to launch variational inference on the model
        specified by code conditioned on data.

        Args
        ----------

        code (string):
            Stan code specified as a string.

        data (dict):
            A Python dictionary containing the data as required
            by the Stan code.

        hparams (dict):
            A dictionary of hyper-parameters choices.
            This dictionary defined the various hyper-parameters
            used throughout the inference scheme. Please take a look
            at the 'algorithm' function for a list of valid attributes.
            A default hyper-parameter dictionary can be found
            at 'vistan/hyperparams.py'

        model_name (string, keyword only argument):
            Default is 'default_model_name'. Alternatively, you can
            specify your own names. Stan does not allow '-' in names.
            So we internally change '-' to '_'.

        verbose (bool, keyword only argument):
            Helps in printing some suppressed model compile warnings.

        print_hparams (bool, keyword only argument):
            Helps print the hyper-parameter dictionary for easier debugging.

        Returns
        -------
        Posterior instance
            A Posterior instance with optimization results, optimizer
            variational parameters, and Model instance locked in.

    """
    npr.seed(hparams['seed'])
    # exit()
    if not code.isascii():
        raise ValueError(f"""Found ascii character in code.
            PyStan currently does not support non-ascii characters.
            See https://github.com/alashworth/stan-monorepo-stage1/issues/87
            for discussion. Please, remove the non-ascii
            characters from the source code.
            {code}""")

    # Stan is particular about model names
    if model_name is not None:
        model_name = model_name.replace("-", "_")

    model = interface.Model(code, data, model_name, verbose=verbose)

    utils.update_hparams(model, hparams)
    if print_hparams:
        print("printing the Hyper-param configuration file...")
        utils.print_hparams(hparams)

    var_dist = vi_families.get_var_dist(hparams)
    init_params = var_dist.initial_params()

    if hparams['LI'] is True:
        init_params = utils.get_LI_params(
                                        hparams=hparams, model=model,
                                        code=code, model_name=model_name)

    objective, eval_function = objectives.get_objective_eval_fn(
                                                    model.log_prob, var_dist,
                                                    hparams)
    objective_grad = autograd.grad(objective)
    optimizer = utils.get_optimizer(hparams)
    callback = utils.get_callback(hparams)
    results, optimized_params = utils.optimization_handler(
                                    objective_grad=objective_grad,
                                    eval_function=eval_function,
                                    init_params=init_params,
                                    step_size=utils.get_step_size(hparams),
                                    optimizer=optimizer,
                                    num_epochs=hparams['max_iters'],
                                    callback=functools.partial(
                                        callback,
                                        model=model,
                                        eval_function=eval_function),
                                    hparams=hparams)
    return vi_families.get_posterior(
                                    model=model, var_dist=var_dist,
                                    M_iw_sample=hparams['M_iw_train'],
                                    params=optimized_params,
                                    results=results)
